from typing import Iterable

import pytest

from mlair.helpers.data_sources.join import *
from mlair.helpers.data_sources.join import _save_to_pandas, _lower_list, _select_distinct_series, \
    _select_distinct_data_origin, _select_distinct_network
from mlair.configuration.join_settings import join_settings
from mlair.helpers.testing import check_nested_equality
from mlair.helpers.data_sources.data_loader import EmptyQueryResult


class TestDownloadJoin:

    def test_download_single_var(self):
        data, meta = download_join("DEBW107", {"o3": "dma8eu"})
        assert data.columns == "o3"
        assert meta.columns == "DEBW107"

    def test_download_empty(self):
        with pytest.raises(EmptyQueryResult) as e:
            download_join("DEBW107", {"o3": "dma8eu"}, "traffic")
        assert e.value.args[-1] == "No data found for variables {'o3'} and options station=['DEBW107'], type=traffic," \
                                   " network=None, origin={} in JOIN."

    def test_download_incomplete(self):
        with pytest.raises(EmptyQueryResult) as e:
            download_join("DEBW107", {"o3": "dma8eu", "o10": "maximum"}, "background")
        assert e.value.args[-1] == "No data found for variables {'o10'} and options station=['DEBW107'], " \
                                   "type=background, network=None, origin={} in JOIN."
        with pytest.raises(EmptyQueryResult) as e:
            download_join("DEBW107", {"o3": "dma8eu", "o10": "maximum"}, "background", data_origin={"o10": ""})
        assert e.value.args[-1] == "No data found for variables {'o10'} and options station=['DEBW107'], " \
                                   "type=background, network={'o10': None}, origin={'o10': ''} in JOIN."


class TestCorrectDataFormat:

    def test_correct_data_format(self):
        list_data = [["2020-01-01 06:00:01", 23.], ["2020-01-01 06:00:11", 24.], ["2020-01-01 06:00:21", 25.],
                     ["2020-01-01 06:00:31", 26.], ["2020-01-01 06:00:41", 27.], ["2020-01-01 06:00:51", 23.],
                     {"station": "test_station_001", "author": "ME", "success": True}]
        dict_data = correct_data_format(list_data)
        assert dict_data == {"datetime": ["2020-01-01 06:00:01", "2020-01-01 06:00:11", "2020-01-01 06:00:21",
                                          "2020-01-01 06:00:31", "2020-01-01 06:00:41", "2020-01-01 06:00:51"],
                             "values": [23., 24., 25., 26., 27., 23.],
                             "metadata": {"station": "test_station_001", "author": "ME", "success": True}}


class TestLoadSeriesInformation:

    def test_standard_query(self):
        expected_subset = {'o3': 17057, 'no2': 17058, 'temp': 85587, 'wspeed': 17060}
        res, orig = load_series_information(['DEBW107'], None, None, join_settings()[0], {})
        assert expected_subset.items() <= res.items()

    def test_empty_result(self):
        res, orig = load_series_information(['DEBW107'], "traffic", None, join_settings()[0], {})
        assert res == {}


class TestSelectDistinctDataOrigin:

    @pytest.fixture
    def vars(self):
        return [{'id': 16686, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'no2',
                 'parameter_label': 'NO2', 'parameter_attribute': ''},
                {'id': 16687, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'o3',
                 'parameter_label': 'O3', 'parameter_attribute': ''},
                {'id': 16692, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'press',
                 'parameter_label': 'PRESS--LANUV', 'parameter_attribute': ''},
                {'id': 16693, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                 'parameter_label': 'TEMP--LANUV', 'parameter_attribute': ''},
                {'id': 54036, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'cloudcover',
                 'parameter_label': 'CLOUDCOVER', 'parameter_attribute': 'REA'},
                {'id': 88491, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                 'parameter_label': 'TEMP-REA-MIUB', 'parameter_attribute': 'REA'},
                {'id': 102660, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'press',
                 'parameter_label': 'PRESS-REA-MIUB', 'parameter_attribute': 'REA'},
                {'id': 26692, 'network_name': 'AIRBASE', 'station_id': 'DENW053', 'parameter_name': 'press',
                 'parameter_label': 'PRESS', 'parameter_attribute': 'REA'}]

    def test_no_origin_given(self, vars):
        res, orig = _select_distinct_data_origin(vars, {})
        expected = {
            "no2": [{'id': 16686, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'no2',
                     'parameter_label': 'NO2', 'parameter_attribute': ''}],
            "o3": [{'id': 16687, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'o3',
                    'parameter_label': 'O3', 'parameter_attribute': ''}],
            "cloudcover": [{'id': 54036, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'cloudcover',
                            'parameter_label': 'CLOUDCOVER', 'parameter_attribute': 'REA'}],
            "temp": [{'id': 88491, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                      'parameter_label': 'TEMP-REA-MIUB', 'parameter_attribute': 'REA'}],
            "press": [{'id': 102660, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'press',
                       'parameter_label': 'PRESS-REA-MIUB', 'parameter_attribute': 'REA'},
                      {'id': 26692, 'network_name': 'AIRBASE', 'station_id': 'DENW053',
                       'parameter_name': 'press', 'parameter_label': 'PRESS', 'parameter_attribute': 'REA'}]}

        assert check_nested_equality(res, expected) is True
        # assert res == {"no2": 16686, "o3": 16687, "cloudcover": 54036, "temp": 88491, "press": 102660}
        assert orig == {"no2": "", "o3": "", "cloudcover": "REA", "temp": "REA", "press": "REA"}

    def test_different_origins(self, vars):
        origin = {"no2": "test", "temp": "", "cloudcover": "REA"}
        res, orig = _select_distinct_data_origin(vars, data_origin=origin)
        expected = {
            "o3": [{'id': 16687, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'o3',
                    'parameter_label': 'O3', 'parameter_attribute': ''}],
            "cloudcover": [{'id': 54036, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'cloudcover',
                            'parameter_label': 'CLOUDCOVER', 'parameter_attribute': 'REA'}],
            "temp": [{'id': 16693, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                      'parameter_label': 'TEMP--LANUV', 'parameter_attribute': ''}],
            "press": [{'id': 102660, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'press',
                       'parameter_label': 'PRESS-REA-MIUB', 'parameter_attribute': 'REA'},
                      {'id': 26692, 'network_name': 'AIRBASE', 'station_id': 'DENW053',
                       'parameter_name': 'press', 'parameter_label': 'PRESS', 'parameter_attribute': 'REA'}]}
        assert check_nested_equality(res, expected) is True
        # assert res == {"o3": 16687, "press": 102660, "temp": 16693, "cloudcover": 54036}
        assert orig == {"no2": "test", "o3": "", "cloudcover": "REA", "temp": "", "press": "REA"}


class TestSelectDistinctNetwork:

    @pytest.fixture
    def vars(self):
        return {
            "no2": [{'id': 16686, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'no2',
                     'parameter_label': 'NO2', 'parameter_attribute': ''}],
            "o3": [{'id': 16687, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'o3',
                    'parameter_label': 'O3', 'parameter_attribute': ''}],
            "cloudcover": [{'id': 54036, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'cloudcover',
                            'parameter_label': 'CLOUDCOVER', 'parameter_attribute': 'REA'}],
            "temp": [{'id': 88491, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                      'parameter_label': 'TEMP-REA-MIUB', 'parameter_attribute': 'REA'}],
            "press": [{'id': 102660, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'press',
                       'parameter_label': 'PRESS-REA-MIUB', 'parameter_attribute': 'REA'},
                      {'id': 26692, 'network_name': 'AIRBASE', 'station_id': 'DENW053',
                       'parameter_name': 'press', 'parameter_label': 'PRESS', 'parameter_attribute': 'REA'}]}

    def test_no_network_given(self, caplog, vars):
        caplog.set_level(logging.INFO)
        res = _select_distinct_network(vars, [])
        expected = {
            "no2": {'id': 16686, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'no2',
                    'parameter_label': 'NO2', 'parameter_attribute': ''},
            "o3": {'id': 16687, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'o3',
                   'parameter_label': 'O3', 'parameter_attribute': ''},
            "cloudcover": {'id': 54036, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'cloudcover',
                           'parameter_label': 'CLOUDCOVER', 'parameter_attribute': 'REA'},
            "temp": {'id': 88491, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                     'parameter_label': 'TEMP-REA-MIUB', 'parameter_attribute': 'REA'},
            "press": {'id': 102660, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'press',
                      'parameter_label': 'PRESS-REA-MIUB', 'parameter_attribute': 'REA'}}
        assert check_nested_equality(res, expected) is True

        message = "Could not find a valid match for variable %s and networks []! Therefore, use first answer from JOIN:"
        assert message % "no2" in caplog.messages[0]
        assert message % "o3" in caplog.messages[1]
        assert message % "cloudcover" in caplog.messages[2]
        assert message % "temp" in caplog.messages[3]
        assert message % "press" in caplog.messages[4]

    def test_single_network_given(self, vars):
        res = _select_distinct_network(vars, ["UBA"])
        expected = {
            "no2": {'id': 16686, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'no2',
                    'parameter_label': 'NO2', 'parameter_attribute': ''},
            "o3": {'id': 16687, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'o3',
                   'parameter_label': 'O3', 'parameter_attribute': ''},
            "cloudcover": {'id': 54036, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'cloudcover',
                           'parameter_label': 'CLOUDCOVER', 'parameter_attribute': 'REA'},
            "temp": {'id': 88491, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                     'parameter_label': 'TEMP-REA-MIUB', 'parameter_attribute': 'REA'},
            "press": {'id': 102660, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'press',
                      'parameter_label': 'PRESS-REA-MIUB', 'parameter_attribute': 'REA'}}
        assert check_nested_equality(res, expected) is True

    def test_single_network_given_no_match(self, vars):
        with pytest.raises(ValueError) as e:  # AIRBASE not avail for all variables
            _select_distinct_network(vars, ["AIRBASE"])
        assert e.value.args[-1] == "Cannot find a valid match for requested networks ['AIRBASE'] and variable no2 as " \
                                   "only following networks are available in JOIN: ['UBA']"

        with pytest.raises(ValueError) as e:  # both requested networks are not available for all variables
            _select_distinct_network(vars, ["LUBW", "EMEP"])
        assert e.value.args[-1] == "Cannot find a valid match for requested networks ['LUBW', 'EMEP'] and variable " \
                                   "no2 as only following networks are available in JOIN: ['UBA']"

    def test_multiple_networks_given(self, vars):
        res = _select_distinct_network(vars, ["UBA", "AIRBASE"])
        expected = {
            "no2": {'id': 16686, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'no2',
                    'parameter_label': 'NO2', 'parameter_attribute': ''},
            "o3": {'id': 16687, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'o3',
                   'parameter_label': 'O3', 'parameter_attribute': ''},
            "cloudcover": {'id': 54036, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'cloudcover',
                           'parameter_label': 'CLOUDCOVER', 'parameter_attribute': 'REA'},
            "temp": {'id': 88491, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                     'parameter_label': 'TEMP-REA-MIUB', 'parameter_attribute': 'REA'},
            "press": {'id': 102660, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'press',
                      'parameter_label': 'PRESS-REA-MIUB', 'parameter_attribute': 'REA'}}
        assert check_nested_equality(res, expected) is True

        res = _select_distinct_network(vars, ["AIRBASE", "UBA"])
        expected = {
            "no2": {'id': 16686, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'no2',
                    'parameter_label': 'NO2', 'parameter_attribute': ''},
            "o3": {'id': 16687, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'o3',
                   'parameter_label': 'O3', 'parameter_attribute': ''},
            "cloudcover": {'id': 54036, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'cloudcover',
                           'parameter_label': 'CLOUDCOVER', 'parameter_attribute': 'REA'},
            "temp": {'id': 88491, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                     'parameter_label': 'TEMP-REA-MIUB', 'parameter_attribute': 'REA'},
            "press": {'id': 26692, 'network_name': 'AIRBASE', 'station_id': 'DENW053',
                      'parameter_name': 'press', 'parameter_label': 'PRESS', 'parameter_attribute': 'REA'}}
        assert check_nested_equality(res, expected) is True

    def test_multiple_networks_given_by_dict(self, vars):
        res = _select_distinct_network(vars, {"no2": "UBA", "o3": ["UBA", "AIRBASE"], "temp": ["AIRBASE", "UBA"],
                                              "press": ["AIRBASE", "UBA"]})
        expected = {
            "no2": {'id': 16686, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'no2',
                    'parameter_label': 'NO2', 'parameter_attribute': ''},
            "o3": {'id': 16687, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'o3',
                   'parameter_label': 'O3', 'parameter_attribute': ''},
            "cloudcover": {'id': 54036, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'cloudcover',
                           'parameter_label': 'CLOUDCOVER', 'parameter_attribute': 'REA'},
            "temp": {'id': 88491, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                     'parameter_label': 'TEMP-REA-MIUB', 'parameter_attribute': 'REA'},
            "press": {'id': 26692, 'network_name': 'AIRBASE', 'station_id': 'DENW053',
                      'parameter_name': 'press', 'parameter_label': 'PRESS', 'parameter_attribute': 'REA'}}
        assert check_nested_equality(res, expected) is True


class TestSelectDistinctSeries:

    @pytest.fixture
    def vars(self):
        return [{'id': 16686, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'no2',
                 'parameter_label': 'NO2', 'parameter_attribute': ''},
                {'id': 16687, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'o3',
                 'parameter_label': 'O3', 'parameter_attribute': ''},
                {'id': 16692, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'press',
                 'parameter_label': 'PRESS--LANUV', 'parameter_attribute': ''},
                {'id': 16693, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                 'parameter_label': 'TEMP--LANUV', 'parameter_attribute': ''},
                {'id': 54036, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'cloudcover',
                 'parameter_label': 'CLOUDCOVER', 'parameter_attribute': 'REA'},
                {'id': 88491, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'temp',
                 'parameter_label': 'TEMP-REA-MIUB', 'parameter_attribute': 'REA'},
                {'id': 102660, 'network_name': 'UBA', 'station_id': 'DENW053', 'parameter_name': 'press',
                 'parameter_label': 'PRESS-REA-MIUB', 'parameter_attribute': 'REA'},
                {'id': 26692, 'network_name': 'AIRBASE', 'station_id': 'DENW053', 'parameter_name': 'press',
                 'parameter_label': 'PRESS', 'parameter_attribute': 'REA'}]

    def test_no_origin_given(self, vars):
        res, orig = _select_distinct_series(vars)
        assert res == {"no2": 16686, "o3": 16687, "cloudcover": 54036, "temp": 88491, "press": 102660}
        assert orig == {"no2": "", "o3": "", "cloudcover": "REA", "temp": "REA", "press": "REA"}

    def test_different_origins(self, vars):
        origin = {"no2": "test", "temp": "", "cloudcover": "REA"}
        res, orig = _select_distinct_series(vars, data_origin=origin)
        assert res == {"o3": 16687, "press": 102660, "temp": 16693, "cloudcover": 54036}
        assert orig == {"no2": "test", "o3": "", "cloudcover": "REA", "temp": "", "press": "REA"}
        res, orig = _select_distinct_series(vars, data_origin={})
        assert res == {"cloudcover": 54036, "no2": 16686, "o3": 16687, "press": 102660, "temp": 88491}
        assert orig == {"no2": "", "o3": "", "temp": "REA", "press": "REA", "cloudcover": "REA"}

    def test_different_networks(self, vars):
        res, orig = _select_distinct_series(vars, network_name="UBA")
        assert res == {"no2": 16686, "o3": 16687, "cloudcover": 54036, "temp": 88491, "press": 102660}
        assert orig == {"no2": "", "o3": "", "cloudcover": "REA", "temp": "REA", "press": "REA"}

        res, orig = _select_distinct_series(vars, network_name=["UBA", "EMEP", "AIRBASE"])
        assert res == {"no2": 16686, "o3": 16687, "cloudcover": 54036, "temp": 88491, "press": 102660}
        assert orig == {"no2": "", "o3": "", "cloudcover": "REA", "temp": "REA", "press": "REA"}

        res, orig = _select_distinct_series(vars, network_name=["EMEP", "AIRBASE", "UBA"])
        assert res == {"no2": 16686, "o3": 16687, "cloudcover": 54036, "temp": 88491, "press": 26692}
        assert orig == {"no2": "", "o3": "", "cloudcover": "REA", "temp": "REA", "press": "REA"}

    def test_network_not_available(self, vars):
        with pytest.raises(ValueError) as e:
            _select_distinct_series(vars, network_name="AIRBASE")
        assert e.value.args[-1] == "Cannot find a valid match for requested networks ['AIRBASE'] and variable no2 as " \
                                   "only following networks are available in JOIN: ['UBA']"

    def test_different_network_and_origin(self, vars):
        origin = {"no2": "test", "temp": "", "cloudcover": "REA"}
        res, orig = _select_distinct_series(vars, data_origin=origin, network_name=["EMEP", "AIRBASE", "UBA"])
        assert res == {"o3": 16687, "press": 26692, "temp": 16693, "cloudcover": 54036}
        assert orig == {"no2": "test", "o3": "", "cloudcover": "REA", "temp": "", "press": "REA"}


class TestSaveToPandas:

    @staticmethod
    def convert_date(date):
        return map(lambda s: dt.datetime.strptime(s, "%Y-%m-%d %H:%M"), date)

    @pytest.fixture
    def date(self):
        return ['1997-01-01 00:00', '1997-01-02 00:00', '1997-01-03 00:00', '1997-01-04 00:00']

    @pytest.fixture
    def date_len19(self):
        return ['1997-01-01 00:00:00', '1997-01-02 00:00:00', '1997-01-03 00:00:00', '1997-01-04 00:00:00']

    @pytest.fixture
    def values(self):
        return [86.21, 94.76, 76.96, 99.89]

    @pytest.fixture
    def alternative_values(self):
        return [20.0, 25.2, 25.1, 23.6]

    @pytest.fixture
    def create_df(self, date, values):
        return pd.DataFrame(values, index=self.convert_date(date), columns=['cloudcover'])

    def test_empty_df(self, date, values, create_df):
        data = {'datetime': date, 'mean': values, 'metadata': None}
        assert pd.testing.assert_frame_equal(create_df, _save_to_pandas(None, data, 'mean', 'cloudcover')) is None

    def test_not_empty_df(self, date, alternative_values, create_df):
        data = {'datetime': date, 'max': alternative_values, 'metadata': None}
        next_df = pd.DataFrame(data["max"], index=self.convert_date(date), columns=['temperature'])
        df_concat = pd.concat([create_df, next_df], axis=1)
        assert pd.testing.assert_frame_equal(df_concat, _save_to_pandas(create_df, data, 'max', 'temperature')) is None

    def test_alternative_date_format(self, date_len19, values, create_df):
        data = {'datetime': date_len19, 'mean': values, 'metadata': None}
        assert pd.testing.assert_frame_equal(create_df, _save_to_pandas(None, data, 'mean', 'cloudcover')) is None


class TestLowerList:

    def test_string_lowering(self):
        list_iterator = _lower_list(["Capitalised", "already_small", "UPPERCASE", "veRyStRaNGe"])
        assert isinstance(list_iterator, Iterable)
        assert list(list_iterator) == ["capitalised", "already_small", "uppercase", "verystrange"]


